package nl.fizzit.maven.plugins.tstamp;

import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.TimeZone;

import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.project.MavenProject;

/**
 * Based on Apache's Ant task TStamp
 * http://ant.apache.org/manual/Tasks/tstamp.html
 * 
 * <pre>
 * &lt;plugin&gt;
 *   &lt;groupId&gt;nl.fizzit.plugins&lt;/groupId&gt;
 *   &lt;artifactId&gt;maven-tstamp-plugin&lt;/artifactId&gt;
 *   &lt;configuration&gt;
 *     &lt;formats&gt;
 *       &lt;format&gt;
 *         &lt;property&gt;test.date&lt;/property&gt;
 *         &lt;pattern&gt;yyyy MMMM dd hhmm&lt;/pattern&gt;
 *         &lt;locale&gt;en,US&lt;/locale&gt;
 *         &lt;timezone&gt;Asia/Tokyo&lt;/timezone&gt;
 *         &lt;offset&gt;-1&lt;/offset&gt;
 *         &lt;unit&gt;hour&lt;/unit&gt;
 *       &lt;/format&gt;
 *       &lt;format&gt;
 *         &lt;property&gt;test2.date&lt;/property&gt;
 *         &lt;pattern&gt;yyyyMMdd&lt;/pattern&gt;
 *         &lt;offset&gt;30&lt;/offset&gt;
 *         &lt;unit&gt;day&lt;/unit&gt;
 *       &lt;/format&gt;
 *     &lt;/formats&gt;
 *   &lt;/configuration&gt;
 * &lt;/plugin&gt;
 * </pre>
 * 
 * Set TStamp properties
 * 
 * @goal export
 * @phase initialize
 */
public class TStampMojo extends AbstractMojo {

	public static final String DSTAMP_PROPERTY_NAME = "DSTAMP";
	public static final String DSTAMP_PROPERTY_FORMAT = "yyyyMMdd";
	public static final String HSTAMP_PROPERTY_NAME = "HSTAMP";
	public static final String HSTAMP_PROPERTY_FORMAT = "hhmm";
	public static final String TODAY_PROPERTY_NAME = "TODAY";
	public static final String TODAY_PROPERTY_FORMAT = "MMMM dd yyyy";
	
	/**
	 * The maven project.
	 * 
	 * @parameter expression="${project}"
	 * @readonly
	 */
	 private MavenProject project;

	
	/**
	 * @parameter
	 */
	private List<Format> formats;

	public void execute() throws MojoExecutionException, MojoFailureException {
		try {
			execute(new Date());
		} catch (Exception e) {
			getLog().debug(e);
			throw new MojoFailureException(String.format("Exception occurred %s: %s", e.getClass().getName(), e.getMessage()));
		}
	}
	
	/**
	 * Method indirectly called by Maven. Introduced for unit testing purposes.
	 * 
	 * @param date
	 * @throws MojoExecutionException
	 * @throws MojoFailureException
	 */
	void execute(Date date) throws MojoExecutionException, MojoFailureException {
		getLog().info("Setting TStamp properties");

		SimpleDateFormat dStampFormatter = new SimpleDateFormat(DSTAMP_PROPERTY_FORMAT);
		exportProperty(DSTAMP_PROPERTY_NAME, dStampFormatter.format(date));

		SimpleDateFormat hStampFormatter = new SimpleDateFormat(HSTAMP_PROPERTY_FORMAT);
		exportProperty(HSTAMP_PROPERTY_NAME, hStampFormatter.format(date));

		SimpleDateFormat todayFormatter = new SimpleDateFormat(TODAY_PROPERTY_FORMAT);
		exportProperty(TODAY_PROPERTY_NAME, todayFormatter.format(date));

		if (formats != null) {
			for (Format format : formats) {
				getLog().info("-- Exporting custom format");
				exportFormattedProperty(format, date);
			}
		}
		getLog().info("TStamp properties set");
	}

	private void exportFormattedProperty(Format format, Date date) {
		// check for required fields
		assertNotNull(format.property);
		assertNotNull(format.pattern);

		TimeZone timezone = TimeZone.getDefault();

		if (format.timezone != null) {
			timezone = TimeZone.getTimeZone(format.timezone);
			// if timezone equals GMT zone, id likely not understood
			getLog().info(String.format("Using timezone: %s", timezone.getDisplayName()));
		}

		Locale locale = Locale.getDefault();

		if (format.locale != null) {
			String[] bits = format.locale.split(",");
			if (bits.length == 1) {
				locale = new Locale(bits[0]);
			} else if (bits.length == 2) {
				locale = new Locale(bits[0], bits[1]);
			} else {
				throw new IllegalArgumentException(String.format("Invalid input for locale: %s", format.locale));
			}
			// if locale equals default one, input likely not fully understood
			getLog().info(String.format("Using locale: %s", locale.getDisplayName()));
		}

		Calendar cal = Calendar.getInstance();
		cal.setTime(date);

		if (format.offset != 0 && format.unit != null) { // or should we use a default unit
			// check unit
			Unit unit = Unit.fromString(format.unit);
			if (unit == null) {
				throw new IllegalArgumentException(String.format("Invalid input for unit: %s", format.unit));
			}

			getLog().info(String.format("Adjusting time offset %s (unit %s)", format.offset, unit.value));
			cal.add(unit.fieldNumber, format.offset);
		}

		SimpleDateFormat customFormatter = new SimpleDateFormat(format.pattern, locale);
		customFormatter.setTimeZone(timezone);
		exportProperty(format.property, customFormatter.format(cal.getTime()));

	}

	private void exportProperty(String key, String value) {
		if (project.getProperties().put(key, value) != null) {
			getLog().warn(String.format("Property key \"%s\" replaced an existing entry.", key));
		}

		getLog().info(String.format("Property %s set to %s", key, value));
	}

	private void assertNotNull(Object value) {
		if (value == null) {
			throw new IllegalArgumentException("Value not allowed to be null");
		}
	}
	
	//not required by Maven but used by unittest
	public void setFormats(List<Format> formats) {
		this.formats = formats;
	}
	
	public void setProject(MavenProject project) {
		this.project = project;
	}
}
